#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <errno.h>

/* romchu 0.2 */
/* an attempt at parsing the version 2 romc */

struct bitstream;

struct bitstream *init_bitstream(const unsigned char *pool, unsigned long pool_size);
uint32_t get_bits(struct bitstream *bs, int bits);
void free_bitstream(struct bitstream *bs);

int main(int argc, char **argv)
{
    FILE *infile = fopen(argv[1], "rb");
    unsigned char head_buf[4];
    unsigned char payload_buf[0x10000];
    int block_count = 0;
    const uint64_t header_mult = 0x400000;
    const uint32_t block_mult = 0x10000;

    uint64_t nominal_size;
    int version;

    {
        /* read header */
        struct bitstream *bs;

        fread(head_buf, 4, 1, infile);

        bs = init_bitstream(head_buf, 4*8);

        nominal_size = get_bits(bs, 24) * header_mult;
        version = get_bits(bs, 8);

        if (version != 2)
        {
            fprintf(stderr,"Expected version 2 romc, got %d\n", version);
            return 1;
        }

        free_bitstream(bs);
    }

    while (1 == fread(head_buf, 4, 1, infile))
    {
        int compression_flag;
        uint32_t payload_bytes;
        int payload_bits;
        uint32_t read_size;

        struct bitstream *head_bs;

#if 1
        printf("%08lx=%08lx\n",
            (unsigned long)(ftell(infile)-4),
            (unsigned long)block_count*block_mult);
#endif

        head_bs = init_bitstream(head_buf, 4*8);

        compression_flag = get_bits(head_bs, 1);
        if (compression_flag)
        {
            /* compressed */

            uint32_t block_size;

            /* bits, including this header */
            block_size = get_bits(head_bs, 31) - 32;

            payload_bytes = block_size/8;
            payload_bits = block_size%8;
        }
        else
        {
            /* uncompressed */

            uint32_t block_size;

            /* bytes */
            block_size = get_bits(head_bs, 31);

            payload_bytes = block_size;
            payload_bits = 0;
        }

        free_bitstream(head_bs);
        head_bs = NULL;

        /* read payload */
        read_size = payload_bytes;
        if (payload_bits > 0)
        {
            read_size ++;
        }

        if (read_size > sizeof(payload_buf))
        {
            fprintf(stderr, "payload too large\n");
            return 1;
        }
        if (1 != fread(payload_buf, read_size, 1, infile))
        {
            int save_errno = errno;
            if (feof(infile))
            {
                fprintf(stderr, "fread of payload: unexpected EOF\n");
            }
            else
            {
                errno = save_errno;
                perror("fread of payload");
            }
            return 1;
        }

        /* attempt to parse... */

        if (compression_flag)
        {
            uint16_t tab1_size, tab2_size;
            uint32_t body_size;
            unsigned long tab1_offset, tab2_offset, body_offset;
            struct bitstream *bs;
            
            /* read table 1 size */
            tab1_offset = 0;
            bs = init_bitstream(payload_buf + tab1_offset, payload_bytes*8+payload_bits);
            tab1_size = get_bits(bs, 16);
            free_bitstream(bs);

            /* check table 1 */
            bs = init_bitstream(payload_buf + tab1_offset + 2, tab1_size);
            free_bitstream(bs);

            /* read table 2 size */
            tab2_offset = tab1_offset + 2 + (tab1_size+7) / 8;
            bs = init_bitstream(payload_buf + tab2_offset, 2*8);
            tab2_size = get_bits(bs, 16);
            free_bitstream(bs);

            /* check table 2 */
            bs = init_bitstream(payload_buf + tab2_offset + 2, tab2_size);
            free_bitstream(bs);

            /* check body */
            body_offset = tab2_offset + 2 + (tab2_size+7) / 8;
            body_size = payload_bytes*8 + payload_bits - body_offset*8;
            bs = init_bitstream(payload_buf + body_offset, body_size);
            free_bitstream(bs);

            printf("%04x %04x %08x\n", (unsigned int)tab1_size, (unsigned int)tab2_size, (unsigned int)body_size);
            //printf("%04x\n", (unsigned int)tab1_size);

#if 0
            /* dump */
            if (tab1_size == 0x735)
            {
                FILE *outfile = fopen("dump.bin", "wb");

                fwrite(payload_buf+tab1_offset+2, (tab1_size+7)/8, 1, outfile);
                fclose(outfile);
            }
#endif

        }

        block_count ++;
    }

    if (block_count * block_mult != nominal_size)
    {
        fprintf(stderr, "size mismatch\n");
        return 1;
    }

    printf("ok!\n");

    return 0;
}

/* bitstream reader */
struct bitstream
{
    const unsigned char *pool;
    long bits_left;
    uint8_t first_byte;
    int first_byte_bits;
};

struct bitstream *init_bitstream(const unsigned char *pool, unsigned long pool_size)
{
    struct bitstream *bb = malloc(sizeof(struct bitstream));
    if (!bb)
    {
        perror("bitstream malloc");
        exit(EXIT_FAILURE);
    }

    bb->pool = pool;
    bb->bits_left = pool_size;
    bb->first_byte_bits = 0;

    /* check that padding bits are 0 (to ensure we aren't ignoring anything) */
    if (pool_size%8)
    {
        if (pool[pool_size/8] & ~((1<<(pool_size%8))-1))
        {
            fprintf(stderr, "nonzero padding at end of bitstream\n");
            exit(EXIT_FAILURE);
        }
    }

    return bb;
}

uint32_t get_bits(struct bitstream *bb, int bits)
{
    uint32_t accum = 0;

    if (bits > 32)
    {
        fprintf(stderr, "get_bits() supports max 32\n");
        exit(EXIT_FAILURE);
    }
    if (bits > bb->bits_left + bb->first_byte_bits)
    {
        fprintf(stderr, "get_bits() underflow\n");
        exit(EXIT_FAILURE);
    }

    for (int i = 0; i < bits; i++)
    {
        if (bb->first_byte_bits == 0)
        {
            bb->first_byte = *bb->pool;
            bb->pool ++;
            if (bb->bits_left >= 8)
            {
                bb->first_byte_bits = 8;
                bb->bits_left -= 8;
            }
            else
            {
                bb->first_byte_bits = bb->bits_left;
                bb->bits_left = 0;
            }
        }

        accum >>= 1;
        accum |= (bb->first_byte & 1)<<31;
        bb->first_byte >>= 1;
        bb->first_byte_bits --;
    }

    return accum>>(32-bits);
}

void free_bitstream(struct bitstream *bb)
{
    free(bb);
}
